# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import warnings

import numpy as np
import pytest

from hyperspy.components1d import Gaussian
from hyperspy.decorators import lazifyTestClass

import exspy
from exspy._defaults_parser import preferences
from exspy.misc.eds import utils as utils_eds
from exspy.signals import EDSTEMSpectrum


@lazifyTestClass
class Test_metadata:
    def setup_method(self, method):
        # Create an empty spectrum
        s = EDSTEMSpectrum(np.ones((4, 2, 1024)))
        s.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time = 3.1
        s.metadata.Acquisition_instrument.TEM.beam_energy = 15.0
        self.signal = s

    def test_sum_minimum_missing(self):
        s = EDSTEMSpectrum(np.ones((4, 2, 1024)))
        s.sum()

    def test_sum_live_time1(self):
        s = self.signal
        old_metadata = s.metadata.deepcopy()
        sSum = s.sum(0)
        assert (
            sSum.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time
            == s.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time * 2
        )
        # Check that metadata is unchanged
        print(old_metadata, s.metadata)  # Capture for comparison on error
        assert (
            old_metadata.as_dictionary() == s.metadata.as_dictionary()
        ), "Source metadata changed"

    def test_sum_live_time2(self):
        s = self.signal
        old_metadata = s.metadata.deepcopy()
        sSum = s.sum((0, 1))
        assert (
            sSum.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time
            == s.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time * 2 * 4
        )
        # Check that metadata is unchanged
        print(old_metadata, s.metadata)  # Capture for comparison on error
        assert (
            old_metadata.as_dictionary() == s.metadata.as_dictionary()
        ), "Source metadata changed"

    def test_sum_live_time_out_arg(self):
        s = self.signal
        sSum = s.sum(0)
        s.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time = 4.2
        s_resum = s.sum(0)
        r = s.sum(0, out=sSum)
        assert r is None
        assert (
            s_resum.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time
            == s.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time * 2
        )
        np.testing.assert_allclose(s_resum.data, sSum.data)

    def test_rebin_live_time(self):
        s = self.signal
        old_metadata = s.metadata.deepcopy()
        dim = s.axes_manager.shape
        s = s.rebin(new_shape=[dim[0] / 2, dim[1] / 2, dim[2]])
        assert (
            s.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time == 3.1 * 2 * 2
        )
        # Check that metadata is unchanged
        print(old_metadata, self.signal.metadata)  # Captured on error
        assert (
            old_metadata.as_dictionary() == self.signal.metadata.as_dictionary()
        ), "Source metadata changed"

    def test_offset_after_rebin(self):
        s = self.signal
        s.axes_manager[0].offset = 1
        s.axes_manager[1].offset = 2
        s.axes_manager[2].offset = 3
        s2 = s.rebin(scale=(2, 2, 1))
        assert s2.axes_manager[0].offset == 1.5
        assert s2.axes_manager[1].offset == 2.5
        assert s2.axes_manager[2].offset == s.axes_manager[2].offset

    def test_add_elements(self):
        s = self.signal
        s.add_elements(["Al", "Ni"])
        assert s.metadata.Sample.elements == ["Al", "Ni"]
        s.add_elements(["Al", "Ni"])
        assert s.metadata.Sample.elements == ["Al", "Ni"]
        s.add_elements(
            [
                "Fe",
            ]
        )
        assert s.metadata.Sample.elements == ["Al", "Fe", "Ni"]
        s.set_elements(["Al", "Ni"])
        assert s.metadata.Sample.elements == ["Al", "Ni"]

    def test_default_param(self):
        s = self.signal
        mp = s.metadata
        assert (
            mp.Acquisition_instrument.TEM.Detector.EDS.energy_resolution_MnKa
            == preferences.EDS.eds_mn_ka
        )

    def test_TEM_to_SEM(self):
        s = self.signal.inav[0, 0]
        signal_type = "EDS_SEM"
        mp = s.metadata.Acquisition_instrument.TEM.Detector.EDS
        mp.energy_resolution_MnKa = 125.3
        sSEM = s.deepcopy()
        sSEM.set_signal_type(signal_type)
        mpSEM = sSEM.metadata.Acquisition_instrument.SEM.Detector.EDS
        results = [mp.energy_resolution_MnKa, signal_type]
        resultsSEM = [mpSEM.energy_resolution_MnKa, sSEM.metadata.Signal.signal_type]
        assert results == resultsSEM

    def test_get_calibration_from(self):
        s = self.signal
        scalib = EDSTEMSpectrum(np.ones(1024))
        energy_axis = scalib.axes_manager.signal_axes[0]
        energy_axis.scale = 0.01
        energy_axis.offset = -0.10
        s.get_calibration_from(scalib)
        assert s.axes_manager.signal_axes[0].scale == energy_axis.scale

    def test_are_microscope_parameters_missing(self):
        assert not self.signal._are_microscope_parameters_missing()

        del self.signal.metadata.Acquisition_instrument.TEM.beam_energy
        del self.signal.metadata.Acquisition_instrument.TEM.Detector.EDS.live_time
        assert self.signal._are_microscope_parameters_missing()


@lazifyTestClass
class Test_quantification:
    def setup_method(self, method):
        s = EDSTEMSpectrum(np.ones([2, 2, 1024]))
        energy_axis = s.axes_manager.signal_axes[0]
        energy_axis.scale = 1e-2
        energy_axis.units = "keV"
        energy_axis.name = "Energy"
        s.set_microscope_parameters(
            beam_energy=200,
            live_time=2.5,
            tilt_stage=0.0,
            azimuth_angle=0,
            elevation_angle=35,
            energy_resolution_MnKa=130,
            beam_current=0.05,
        )
        elements = ["Al", "Zn"]
        xray_lines = ["Al_Ka", "Zn_Ka"]
        intensities = [300, 500]
        for i, xray_line in enumerate(xray_lines):
            gauss = Gaussian()
            line_energy, FWHM = s._get_line_energy(xray_line, FWHM_MnKa="auto")
            gauss.centre.value = line_energy
            gauss.A.value = intensities[i]
            gauss.sigma.value = FWHM
            s.data[:] += gauss.function(energy_axis.axis)

        s.set_elements(elements)
        s.add_lines(xray_lines)
        s.axes_manager[0].scale = 0.5
        s.axes_manager[1].scale = 0.5
        s.axes_manager[0].units = "nm"
        s.axes_manager[1].units = "nm"
        self.signal = s

    def test_metadata(self):
        TEM_md = self.signal.metadata.Acquisition_instrument.TEM
        np.testing.assert_approx_equal(TEM_md.beam_energy, 200)
        np.testing.assert_approx_equal(TEM_md.beam_current, 0.05)
        np.testing.assert_approx_equal(TEM_md.Stage.tilt_alpha, 0.0)
        np.testing.assert_approx_equal(TEM_md.Detector.EDS.live_time, 2.5)
        np.testing.assert_approx_equal(TEM_md.Detector.EDS.elevation_angle, 35)
        np.testing.assert_approx_equal(TEM_md.Detector.EDS.energy_resolution_MnKa, 130)

        self.signal.set_microscope_parameters(real_time=3.1)
        self.signal.set_microscope_parameters(probe_area=1.2)
        np.testing.assert_approx_equal(TEM_md.probe_area, 1.2)
        np.testing.assert_approx_equal(TEM_md.Detector.EDS.real_time, 3.1)

    def test_quant_one_intensity_error(self):
        s = self.signal
        method = "CL"
        kfactors = [1, 2.0009344042484134]
        intensities = s.get_lines_intensity()[:1]
        assert len(intensities) == 1
        with pytest.raises(ValueError):
            _ = s.quantification(intensities, method, kfactors)

        intensities = s.get_lines_intensity()[0]
        with pytest.raises(ValueError):
            _ = s.quantification(intensities, method, kfactors)

    def test_quant_lorimer(self):
        s = self.signal
        method = "CL"
        kfactors = [1, 2.0009344042484134]
        composition_units = "weight"
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, kfactors, composition_units)
        s2 = s.rebin(new_shape=(1, 1, 1024)).squeeze()
        s2.quantification(
            intensities, method, kfactors, composition_units, plot_result=True
        )
        np.testing.assert_allclose(res[0].data, np.ones((2, 2)) * 22.70779, atol=1e-3)

        # Test plot_results
        s2 = s.inav[0, 0]
        s2.quantification(
            intensities, method, kfactors, composition_units, plot_result=True
        )
        np.testing.assert_allclose(res[0].data, 22.70779)

    def test_quant_lorimer_mask(self):
        s = self.signal
        method = "CL"
        kfactors = [1, 2.0009344042484134]
        composition_units = "weight"
        intensities = s.get_lines_intensity()
        mask = np.array([[1, 1], [0, 0]])
        res = s.quantification(
            intensities, method, kfactors, composition_units, navigation_mask=mask
        )
        np.testing.assert_allclose(
            res[0].data, np.array([[0, 0], [22.70779, 22.70779]]), atol=1e-3
        )

    def test_quant_lorimer_warning(self):
        s = self.signal
        method = "CL"
        kfactors = [1, 2.0009344042484134]
        composition_units = "weight"
        intensities = s.get_lines_intensity()
        with pytest.raises(ValueError, match="Thickness is required for absorption"):
            _ = s.quantification(
                intensities,
                method,
                kfactors,
                composition_units,
                absorption_correction=True,
                thickness=None,
            )

    def test_quant_lorimer_ac(self):
        s = self.signal
        method = "CL"
        kfactors = [1, 2.0009344042484134]
        composition_units = "weight"
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, kfactors, composition_units)
        np.testing.assert_allclose(res[0].data, np.ones((2, 2)) * 22.70779, atol=1e-3)
        res2 = s.quantification(
            intensities,
            method,
            kfactors,
            composition_units,
            absorption_correction=True,
            thickness=1.0,
        )
        res3 = s.quantification(
            intensities,
            method,
            kfactors,
            composition_units,
            absorption_correction=True,
            thickness=300.0,
        )
        res4 = s.quantification(
            intensities,
            method,
            kfactors,
            composition_units,
            absorption_correction=True,
            thickness=0.0001,
        )
        list.reverse(intensities)
        list.reverse(kfactors)
        res5 = s.quantification(
            intensities,
            method,
            kfactors,
            composition_units,
            absorption_correction=True,
            thickness=300.0,
        )
        np.testing.assert_allclose(res5[0][0].data, res3[0][1].data, atol=1e-5)
        np.testing.assert_allclose(
            res2[0][0].data, np.ones((2, 2)) * 22.743013, atol=1e-3
        )
        np.testing.assert_allclose(
            res3[0][0].data, np.ones((2, 2)) * 31.816908, atol=1e-3
        )
        np.testing.assert_allclose(res[0].data, res4[0][0].data, atol=1e-5)

    def test_quant_zeta(self):
        s = self.signal
        method = "zeta"
        composition_units = "weight"
        factors = [20, 50]
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, factors, composition_units)
        np.testing.assert_allclose(
            res[1].data, np.ones((2, 2)) * 2.7125736e-03, atol=1e-3
        )
        np.testing.assert_allclose(
            res[0][1].data, np.ones((2, 2)) * 80.962287987, atol=1e-3
        )
        res2 = s.quantification(
            intensities,
            method,
            factors,
            composition_units,
            absorption_correction=True,
            thickness=1.0,
        )
        res3 = s.quantification(
            intensities,
            method,
            factors,
            composition_units,
            absorption_correction=True,
            thickness=100.0,
        )
        assert res2 == res3
        np.testing.assert_allclose(
            res2[0][1].data, np.ones((2, 2)) * 65.5867, atol=1e-3
        )

    def test_quant_cross_section_units(self):
        s = self.signal.deepcopy()
        s2 = self.signal.deepcopy()
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].units = "µm"
        s.axes_manager[0].scale = 0.5 / 1000
        s.axes_manager[1].scale = 0.5 / 1000

        method = "cross_section"
        factors = [3, 5]
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, factors)
        res2 = s2.quantification(intensities, method, factors)
        np.testing.assert_allclose(res[0][0].data, res2[0][0].data)
        # Check that the quantification doesn't change the units of the signal
        assert s.axes_manager[0].units == "µm"
        assert s.axes_manager[1].units == "µm"

    @pytest.mark.parametrize("axes", (None, "nav_axes", [0, 1], ["x", "y"]))
    def test_get_probe_area(self, axes):
        s = self.signal
        s.axes_manager[0].name = "x"
        s.axes_manager[1].name = "y"
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].units = "µm"
        s.axes_manager[0].scale = 0.5 / 1000
        s.axes_manager[1].scale = 0.5 / 1000
        if axes == "nav_axes":
            axes = s.axes_manager.navigation_axes
        np.testing.assert_allclose(s.get_probe_area(axes), 0.25, atol=1e-3)

    @pytest.mark.parametrize("axes", (None, "nav_axes", [0], ["x"]))
    def test_get_probe_area_line_scan(self, axes):
        s = self.signal.inav[0]
        s.axes_manager[0].name = "x"
        s.axes_manager[0].units = "µm"
        s.axes_manager[0].scale = 0.5 / 1000
        if axes == "nav_axes":
            axes = s.axes_manager.navigation_axes
        np.testing.assert_allclose(s.get_probe_area(axes), 0.25, atol=1e-3)

    @pytest.mark.parametrize("axes", (None, "nav_axes", [0], ["x"]))
    def test_get_probe_area_line_scan_other_nav_axes(self, axes):
        s = self.signal
        s.axes_manager[0].name = "x"
        s.axes_manager[1].name = "time"
        s.axes_manager[0].units = "µm"
        s.axes_manager[1].units = "s"
        s.axes_manager[0].scale = 0.5 / 1000
        s.axes_manager[1].scale = 10
        if axes == "nav_axes" or axes is None:
            axes = s.axes_manager.navigation_axes
            with pytest.raises(ValueError):
                s.get_probe_area(axes)
        else:
            np.testing.assert_allclose(s.get_probe_area(axes), 0.25, atol=1e-3)

    def test_zeta_vs_cross_section(self):
        s = self.signal
        factors = [3, 5]
        method = "zeta"
        intensities = s.get_lines_intensity()
        zfactors = utils_eds.edx_cross_section_to_zeta([3, 5], ["Al", "Zn"])
        factors2 = utils_eds.zeta_to_edx_cross_section(zfactors, ["Al", "Zn"])
        np.testing.assert_allclose(factors, factors2, atol=1e-3)

        res = s.quantification(
            intensities,
            method,
            factors=utils_eds.edx_cross_section_to_zeta(
                [22.402, 21.7132], ["Al", "Zn"]
            ),
        )
        res2 = s.quantification(
            intensities, method="cross_section", factors=[22.402, 21.7132]
        )
        np.testing.assert_allclose(res[0][0].data, res2[0][0].data, atol=1e-3)
        np.testing.assert_allclose(res[0][0].data, np.ones((2, 2)) * 36.2969, atol=1e-3)

    def test_quant_cross_section(self):
        s = self.signal
        method = "cross_section"
        factors = [3, 5]
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, factors)
        np.testing.assert_allclose(
            res[1][0].data, np.ones((2, 2)) * 21517.1647, atol=1e-3
        )
        np.testing.assert_allclose(
            res[1][1].data, np.ones((2, 2)) * 21961.6166, atol=1e-3
        )
        np.testing.assert_allclose(res[0][0].data, np.ones((2, 2)) * 49.4889, atol=1e-3)

    def test_method_error(self):
        s = self.signal
        method = "random_method"
        factors = [3, 5]
        intensities = s.get_lines_intensity()
        with pytest.raises(
            ValueError, match="Please specify method for quantification"
        ):
            _ = s.quantification(intensities, method, factors)

    def test_quant_cross_section_ac(self):
        s = self.signal
        method = "cross_section"
        zfactors = [20, 50]
        factors = utils_eds.zeta_to_edx_cross_section(zfactors, ["Al", "Zn"])
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, factors, absorption_correction=True)
        res2 = s.quantification(
            intensities, method="zeta", factors=zfactors, absorption_correction=True
        )
        np.testing.assert_allclose(
            res[0][1].data, np.ones((2, 2)) * 44.02534, atol=1e-3
        )
        np.testing.assert_allclose(res2[0][0].data, res[0][0].data, atol=1e-3)

    def test_quant_zeros(self):
        intens = np.array(
            [
                [0.5, 0.5, 0.5],
                [0.0, 0.5, 0.5],
                [0.5, 0.0, 0.5],
                [0.5, 0.5, 0.0],
                [0.5, 0.0, 0.0],
            ]
        ).T
        with warnings.catch_warnings():
            warnings.filterwarnings(
                "ignore", message="divide by zero encountered", category=RuntimeWarning
            )
            quant = utils_eds.quantification_cliff_lorimer(intens, [1, 1, 3]).T
        np.testing.assert_allclose(
            quant,
            np.array(
                [
                    [0.2, 0.2, 0.6],
                    [0.0, 0.25, 0.75],
                    [0.25, 0.0, 0.75],
                    [0.5, 0.5, 0.0],
                    [1.0, 0.0, 0.0],
                ]
            ),
        )

    def test_edx_cross_section_to_zeta(self):
        cs = [3, 6]
        elements = ["Pt", "Ni"]
        res = utils_eds.edx_cross_section_to_zeta(cs, elements)
        np.testing.assert_allclose(res, [1079.815272, 162.4378035], atol=1e-3)

    def test_zeta_to_edx_cross_section(self):
        factors = [1079.815272, 162.4378035]
        elements = ["Pt", "Ni"]
        res = utils_eds.zeta_to_edx_cross_section(factors, elements)
        np.testing.assert_allclose(res, [3, 6], atol=1e-3)

    def test_quant_element_order(self):
        s = self.signal
        s.set_elements([])
        s.set_lines([])
        lines = ["Zn_Ka", "Al_Ka"]
        kfactors = [2.0009344042484134, 1]
        intensities = s.get_lines_intensity(xray_lines=lines)
        res = s.quantification(
            intensities, method="CL", factors=kfactors, composition_units="weight"
        )
        assert res[0].metadata.Sample.xray_lines[0] == "Zn_Ka"
        assert res[1].metadata.Sample.xray_lines[0] == "Al_Ka"
        np.testing.assert_allclose(res[1].data, np.ones((2, 2)) * 22.70779, atol=1e-3)

    def test_CL_get_mass_thickness(self):
        s = self.signal
        method = "CL"
        kfactors = [1, 2.0009344042484134]
        composition_units = "weight"
        intensities = s.get_lines_intensity()
        res = s.quantification(intensities, method, kfactors, composition_units)

        mass_thickness = s.CL_get_mass_thickness(res, 100.0)[0, 0]
        np.testing.assert_allclose(mass_thickness, 6.1317741e-4)

        thickness = np.array([[100.0, 90.0], [85, 80.0]])
        mass_thickness2 = s.CL_get_mass_thickness(res, thickness)
        np.testing.assert_allclose(mass_thickness2, mass_thickness * thickness / 100)


@lazifyTestClass
class Test_vacuum_mask:
    def setup_method(self, method):
        s = EDSTEMSpectrum(np.array([np.linspace(0.001, 0.5, 20)] * 100).T)
        s.add_poissonian_noise()
        self.signal = s

    def test_vacuum_mask(self):
        s = self.signal
        assert s.vacuum_mask().data[0]
        assert not s.vacuum_mask().data[-1]

    def test_vacuum_mask_navigation_dimension_0(self):
        s = self.signal
        s2 = s.sum()
        with pytest.raises(RuntimeError):
            s2.vacuum_mask()


@pytest.mark.parametrize("normalise_poissonian_noise", [True, False])
def test_decomposition(normalise_poissonian_noise):
    s = EDSTEMSpectrum(np.ones(shape=(32, 32, 1024)))
    s.add_poissonian_noise()
    # default uses `vacuum_mask`
    s.decomposition(normalise_poissonian_noise)

    # test with numpy array mask
    mask = s.vacuum_mask().data
    s.decomposition(normalise_poissonian_noise, navigation_mask=mask)


@lazifyTestClass
class Test_simple_model:
    def setup_method(self, method):
        s = utils_eds.xray_lines_model(elements=["Al", "Zn"], weight_percents=[50, 50])
        self.signal = s

    def test_intensity(self):
        s = self.signal
        np.testing.assert_allclose(
            [i.data[0] for i in s.get_lines_intensity(integration_window_factor=5.0)],
            [0.5, 0.5],
            atol=1e-1,
        )

    def test_intensity_dtype_uint(self):
        s = self.signal
        s.data *= 1e5
        s.change_dtype("uint")
        bw = s.estimate_background_windows()
        np.testing.assert_allclose(
            [i.data[0] for i in s.get_lines_intensity(background_windows=bw)],
            [5e4, 5e4],
            rtol=0.03,
        )


def test_with_signals_examples():
    sig = exspy.data.EDS_TEM_FePt_nanoparticles()
    for s in (sig, sig.as_lazy()):
        np.testing.assert_allclose(
            np.array([res.data[0] for res in s.get_lines_intensity()]),
            np.array([3710, 15872]),
        )


class Test_eds_markers:
    def setup_method(self, method):
        s = utils_eds.xray_lines_model(elements=["Al", "Zn"], weight_percents=[50, 50])
        self.signal = s

    def test_plot_auto_add(self):
        s = self.signal
        s.plot(xray_lines=True)
        # Should contain 6 lines
        assert sorted(
            s._xray_markers["names"]
            == ["Al_Ka", "Al_Kb", "Zn_Ka", "Zn_Kb", "Zn_La", "Zn_Lb1"]
        )

    def test_manual_add_line(self):
        s = self.signal
        s.plot()
        s.add_xray_lines_markers(["Zn_La"])
        assert list(s._xray_markers["names"]) == ["Zn_La"]
        assert len(s._xray_markers["names"]) == 1

    def test_manual_add_line_error(self):
        s = self.signal
        with pytest.raises(RuntimeError):
            s.add_xray_lines_markers(["Zn_La"])

    def test_manual_remove_element(self):
        s = self.signal
        s.plot()
        s.add_xray_lines_markers(["Zn_Ka", "Zn_Kb", "Zn_La"])
        s.remove_xray_lines_markers(["Zn_Kb"])
        assert sorted(s._xray_markers["names"]) == ["Zn_Ka", "Zn_La"]
        s.remove_xray_lines_markers(["Zn_Ka"], render_figure=False)
        assert sorted(s._xray_markers["names"]) == ["Zn_La"]
        s.remove_xray_lines_markers(["Zn_La"], render_figure=True)
        assert sorted(s._xray_markers["names"]) == []
        assert len(s._xray_markers["texts"].get_current_kwargs()["offsets"]) == 0
        assert len(s._xray_markers["texts"].get_current_kwargs()["texts"]) == 0
        assert len(s._xray_markers["lines"].get_current_kwargs()["segments"]) == 0



"""
Plot Residual
=============

Fit an affine function and plot the residual.

"""

import numpy as np
import hyperspy.api as hs

# %%
# Create a signal:
data = np.arange(1000, dtype=np.int64).reshape((10, 100))
s = hs.signals.Signal1D(data)

# %%
# Add noise:
s.add_poissonian_noise(random_state=0)

# %%
# Create model:
m = s.create_model()
line = hs.model.components1D.Expression("a * x + b", name="Affine")
m.append(line)

# %%
# Fit for all navigation positions:
m.multifit()

# %%
# Plot the fitted model with residual:
m.plot(plot_residual=True)

# %%
# sphinx_gallery_thumbnail_number = 2



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import warnings

import pytest
import requests
from requests.exceptions import SSLError

from exspy.misc.eels.eelsdb import eelsdb


def _eelsdb(**kwargs):
    try:
        ss = eelsdb(**kwargs)
    except SSLError:
        warnings.warn(
            "The https://eelsdb.eu certificate seems to be invalid. "
            "Consider notifying the issue to the EELSdb webmaster."
        )
        ss = eelsdb(verify_certificate=False, **kwargs)
    except Exception as e:
        # e.g. failures such as ConnectionError or MaxRetryError
        pytest.skip(f"Skipping eelsdb test due to {e}")
    return ss


def eelsdb_down():
    try:
        _ = requests.get("https://api.eelsdb.eu", verify=True)
        return False
    except SSLError:
        _ = requests.get("https://api.eelsdb.eu", verify=False)
        return False
    except requests.exceptions.ConnectionError:
        return True


@pytest.mark.skipif(eelsdb_down(), reason="Unable to connect to EELSdb")
def test_eelsdb_eels():
    ss = _eelsdb(
        title="Boron Nitride Multiwall Nanotube",
        formula="BN",
        spectrum_type="coreloss",
        edge="K",
        min_energy=370,
        max_energy=1000,
        min_energy_compare="gt",
        max_energy_compare="lt",
        resolution="0.7 eV",
        resolution_compare="lt",
        max_n=2,
        order="spectrumMin",
        order_direction="DESC",
        monochromated=False,
    )

    assert len(ss) == 2
    md = ss[0].metadata
    assert md.General.author == "Odile Stephan"
    assert md.Acquisition_instrument.TEM.Detector.EELS.collection_angle == 24
    assert md.Acquisition_instrument.TEM.convergence_angle == 15
    assert md.Acquisition_instrument.TEM.beam_energy == 100
    assert md.Signal.signal_type == "EELS"
    assert "perpendicular" in md.Sample.description
    assert "parallel" in ss[1].metadata.Sample.description
    assert md.Sample.chemical_formula == "BN"
    assert md.Acquisition_instrument.TEM.microscope == "STEM-VG"


@pytest.mark.skipif(eelsdb_down(), reason="Unable to connect to EELSdb")
def test_eelsdb_xas():
    ss = _eelsdb(
        spectrum_type="xrayabs",
        max_n=1,
    )
    assert len(ss) == 1
    md = ss[0].metadata
    assert md.Signal.signal_type == "XAS"


@pytest.mark.skipif(eelsdb_down(), reason="Unable to connect to EELSdb")
def test_eelsdb_corrupted_file():
    ss = _eelsdb(spectrum_type="coreloss", element="Cu", edge="K", formula="Cu4O3")
    assert len(ss) == 1
    assert ss[0].metadata.General.title == "O-K edge in Cu4O3"


@pytest.mark.skipif(eelsdb_down(), reason="Unable to connect to EELSdb")
def test_eelsdb_elements_no():
    title = "Zero-loss c-FEG Hitachi Disp 0.214 eV"
    ss = _eelsdb(author="Luc Lajaunie", title=title)
    assert len(ss) == 1
    assert ss[0].metadata.General.title == title


@pytest.mark.skipif(eelsdb_down(), reason="Unable to connect to EELSdb")
def test_eelsdb_txt_file():
    _ = _eelsdb(title="Porous cobalt coating with He-filled nanopores")



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

try:
    # Set traits toolkit to work in a headless system
    # Capture error when toolkit is already previously set which typically
    # occurs when building the doc locally
    from traits.etsconfig.api import ETSConfig

    ETSConfig.toolkit = "null"
except ValueError:
    # in case ETSConfig.toolkit was already set previously.
    pass

# pytest-mpl 0.7 already import pyplot, so setting the matplotlib backend to
# 'agg' as early as we can is useless for testing.
import matplotlib.pyplot as plt

import pytest
import numpy as np
import matplotlib
import hyperspy.api as hs


@pytest.fixture(autouse=True)
def add_np(doctest_namespace):
    doctest_namespace["np"] = np
    doctest_namespace["plt"] = plt
    doctest_namespace["hs"] = hs


@pytest.fixture
def pdb_cmdopt(request):
    return request.config.getoption("--pdb")


def setup_module(mod, pdb_cmdopt):
    if pdb_cmdopt:
        import dask

        dask.set_options(get=dask.local.get_sync)


from matplotlib.testing.conftest import mpl_test_settings


try:
    import pytest_mpl
except ImportError:
    # Register dummy marker to allow running the test suite without pytest-mpl
    def pytest_configure(config):
        config.addinivalue_line(
            "markers",
            "mpl_image_compare: dummy marker registration to allow running "
            "without the pytest-mpl plugin.",
        )



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
import logging

from hyperspy._components.expression import Expression

_logger = logging.getLogger(__name__)


class SEE(Expression):
    r"""Secondary electron emission component for Photoemission Spectroscopy.

    .. math::
        :nowrap:

        \[
        f(x) =
        \begin{cases}
            0, & x \leq \Phi\\
            A\cdot{ (x-\Phi) / (x-\Phi+B)^{4}}, & x >  \Phi
        \end{cases}
        \]

    ============= =============
     Variable      Parameter
    ============= =============
     :math:`A`     A
     :math:`\Phi`  Phi
     :math:`B`     B
    ============= =============

    Parameters
    ----------
    A : float
        Height parameter
    Phi : float
        Position parameter
    B : float
        Tail or asymmetry parameter
    **kwargs
        Extra keyword arguments are passed to the
        :py:class:`~._components.expression.Expression` component.

    """

    def __init__(
        self, A=1.0, Phi=1.0, B=0.0, module="numexpr", compute_gradients=False, **kwargs
    ):
        if kwargs.pop("sigma", False):
            _logger.warning(
                "The `sigma` parameter was broken and it has been " "removed."
            )

        super().__init__(
            expression="where(x > Phi, A * (x - Phi) / (x - Phi + B) ** 4, 0)",
            name="SEE",
            A=A,
            Phi=Phi,
            B=B,
            position="Phi",
            module=module,
            autodoc=False,
            compute_gradients=compute_gradients,
            linear_parameter_list=["A"],
            check_parameter_linearity=False,
            **kwargs,
        )

        # Boundaries
        self.A.bmin = 0.0
        self.A.bmax = None

        self.convolved = True

    def grad_A(self, x):
        """ """
        return np.where(
            x > self.Phi.value,
            (x - self.Phi.value) / (x - self.Phi.value + self.B.value) ** 4,
            0,
        )

    def grad_Phi(self, x):
        """ """
        return np.where(
            x > self.Phi.value,
            (4 * (x - self.Phi.value) * self.A.value)
            / (self.B.value + x - self.Phi.value) ** 5
            - self.A.value / (self.B.value + x - self.Phi.value) ** 4,
            0,
        )

    def grad_B(self, x):
        return np.where(
            x > self.Phi.value,
            -(4 * (x - self.Phi.value) * self.A.value)
            / (self.B.value + x - self.Phi.value) ** 5,
            0,
        )



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

from exspy.misc.eds.utils import get_xray_lines_near_energy, take_off_angle


def test_xray_lines_near_energy():
    E = 1.36
    lines = get_xray_lines_near_energy(E)
    assert lines == [
        "Pm_M2N4",
        "Ho_Ma",
        "Eu_Mg",
        "Se_La",
        "Br_Ln",
        "W_Mz",
        "As_Lb3",
        "Kr_Ll",
        "Ho_Mb",
        "Ta_Mz",
        "Dy_Mb",
        "As_Lb1",
        "Gd_Mg",
        "Er_Ma",
        "Sm_M2N4",
        "Mg_Kb",
        "Se_Lb1",
        "Ge_Lb3",
        "Br_Ll",
        "Sm_Mg",
        "Dy_Ma",
        "Nd_M2N4",
        "As_La",
        "Re_Mz",
        "Hf_Mz",
        "Kr_Ln",
        "Er_Mb",
        "Tb_Mb",
    ]
    lines = get_xray_lines_near_energy(E, 0.02)
    assert lines == ["Pm_M2N4"]
    E = 5.4
    lines = get_xray_lines_near_energy(E)
    assert lines == [
        "Cr_Ka",
        "La_Lb2",
        "V_Kb",
        "Pm_La",
        "Pm_Ln",
        "Ce_Lb3",
        "Gd_Ll",
        "Pr_Lb1",
        "Xe_Lg3",
        "Pr_Lb4",
    ]
    lines = get_xray_lines_near_energy(E, only_lines=("a", "b"))
    assert lines == ["Cr_Ka", "V_Kb", "Pm_La", "Pr_Lb1"]
    lines = get_xray_lines_near_energy(E, only_lines=("a"))
    assert lines == ["Cr_Ka", "Pm_La"]


def test_takeoff_angle():
    np.testing.assert_allclose(40.0, take_off_angle(30.0, 0.0, 10.0))
    np.testing.assert_allclose(40.0, take_off_angle(0.0, 90.0, 10.0, beta_tilt=30.0))
    np.testing.assert_allclose(
        73.15788376370121, take_off_angle(45.0, 45.0, 45.0, 45.0)
    )



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import copy
import logging
import warnings

import numpy as np

from hyperspy import components1d
from exspy.signals.eels import EELSSpectrum
from exspy.components import EELSCLEdge
from hyperspy.components1d import PowerLaw
from hyperspy.docstrings.model import FIT_PARAMETERS_ARG
from exspy.docstrings.model import EELSMODEL_PARAMETERS
from hyperspy.misc.utils import dummy_context_manager
from hyperspy.models.model1d import Model1D


_logger = logging.getLogger(__name__)


def generate_uniform_axis(offset, scale, size, offset_index=0):
    """Creates a uniform axis vector given the offset, scale and number of
    channels.

    Alternatively, the offset_index of the offset channel can be specified.

    Parameters
    ----------
    offset : float
    scale : float
    size : number of channels
    offset_index : int
        offset_index number of the offset

    Returns
    -------
    Numpy array

    """

    return np.linspace(
        offset - offset_index * scale, offset + scale * (size - 1 - offset_index), size
    )


class EELSModel(Model1D):
    def __init__(
        self,
        signal1D,
        auto_background=True,
        auto_add_edges=True,
        low_loss=None,
        GOS="gosh",
        dictionary=None,
    ):
        """
        Build an EELS model.

        Parameters
        ----------
        spectrum : a EELSSpectrum instance
        %s

        """
        super().__init__(signal1D)

        # When automatically setting the fine structure energy regions,
        # the fine structure of an EELS edge component is automatically
        # disable if the next ionisation edge onset distance to the
        # higher energy side of the fine structure region is lower that
        # the value of this parameter
        self._min_distance_between_edges_for_fine_structure = 0
        self._preedge_safe_window_width = 2
        self._suspend_auto_fine_structure_width = False
        self._low_loss = None
        self._convolved = False
        self.convolution_axis = None
        self.low_loss = low_loss
        self.GOS = GOS
        self.edges = []
        self._background_components = []
        self._whitelist.update(
            {
                "_convolved": None,
                "low_loss": ("sig", None),
            }
        )
        self._slicing_whitelist["low_loss"] = "inav"
        if dictionary is not None:
            auto_background = False
            auto_add_edges = False
            self._load_dictionary(dictionary)
            for edge in self.edges:
                fine_structure_components = set()
                for comp_name in edge.fine_structure_components:
                    fine_structure_components.add(self[comp_name])
                edge.fine_structure_components = fine_structure_components

        if auto_background is True:
            background = PowerLaw()
            self.append(background)

        if self.signal.subshells and auto_add_edges is True:
            self._add_edges_from_subshells_names()

    __init__.__doc__ %= EELSMODEL_PARAMETERS

    def _compute_constant_term(self, component):
        """Gets the value of any (non-free) constant term, with convolution"""
        if self.convolved and component.convolved:
            data = self._convolve_component_values(component._constant_term)
            return data.T[np.where(self._channel_switches)[::-1]].T
        else:
            return super()._compute_constant_term(component)

    def _convolve_component_values(self, component_values):
        """
        Convolve component with model convolution axis.

        Multiply by np.ones in order to handle case where component_values is a
        single constant
        """

        sig = component_values * np.ones(self.convolution_axis.shape)

        ll = self.low_loss._get_current_data(self.axes_manager)
        convolved = np.convolve(sig, ll, mode="valid")

        return convolved

    def _get_model_data(self, *args, **kwargs):
        if self.convolved is False:
            return super()._get_model_data(*args, **kwargs)
        else:  # convolved
            component_list = kwargs.get("component_list")
            ignore_channel_switches = kwargs.get("ignore_channel_switches", False)
            slice_ = slice(None) if ignore_channel_switches else self._channel_switches
            if self.convolution_axis is None:
                raise RuntimeError("`low_loss` is not set.")
            sum_convolved = np.zeros(len(self.convolution_axis))
            sum_ = np.zeros(len(self.axis.axis))
            for component in component_list:
                if component.convolved:
                    sum_convolved += component.function(self.convolution_axis)
                else:
                    sum_ += component.function(self.axis.axis)
            to_return = sum_ + np.convolve(
                self.low_loss._get_current_data(self.axes_manager),
                sum_convolved,
                mode="valid",
            )
            to_return = to_return[slice_]
            return to_return

    def _jacobian(self, param, y, weights=None):
        if not self.convolved:
            return super()._jacobian(param, y, weights)

        if weights is None:
            weights = 1.0

        counter = 0
        grad = np.zeros(len(self.axis.axis))
        for component in self:  # Cut the parameters list
            if component.active:
                component.fetch_values_from_array(
                    param[counter : counter + component._nfree_param], onlyfree=True
                )

                if component.convolved:
                    for parameter in component.free_parameters:
                        par_grad = np.convolve(
                            parameter.grad(self.convolution_axis),
                            self.low_loss._get_current_data(self.axes_manager),
                            mode="valid",
                        )

                        if parameter._twins:
                            for par in parameter._twins:
                                np.add(
                                    par_grad,
                                    np.convolve(
                                        par.grad(self.convolution_axis),
                                        self.low_loss._get_current_data(
                                            self.axes_manager
                                        ),
                                        mode="valid",
                                    ),
                                    par_grad,
                                )

                        grad = np.vstack((grad, par_grad))

                else:
                    for parameter in component.free_parameters:
                        par_grad = parameter.grad(self.axis.axis)

                        if parameter._twins:
                            for par in parameter._twins:
                                np.add(par_grad, par.grad(self.axis.axis), par_grad)

                        grad = np.vstack((grad, par_grad))

                counter += component._nfree_param

        to_return = grad[1:, self._channel_switches] * weights

        if self.axis.is_binned:
            if self.axis.is_uniform:
                to_return *= self.axis.scale
            else:
                to_return *= np.gradient(self.axis.axis)

        return to_return

    @property
    def signal(self):
        return self._signal

    @signal.setter
    def signal(self, value):
        if isinstance(value, EELSSpectrum):
            self._signal = value
        else:
            raise ValueError(
                "This attribute can only contain an EELSSpectrum "
                "but an object of type %s was provided" % str(type(value))
            )

    @property
    def convolved(self):
        return self._convolved

    @convolved.setter
    def convolved(self, value):
        if isinstance(value, bool):
            if value is not self._convolved:
                if value and not self.low_loss:
                    raise RuntimeError(
                        "Cannot set `convolved` to True as the "
                        "`low_loss` attribute"
                        "is not set."
                    )
                else:
                    self._convolved = value
                    self.update_plot()
        else:
            raise ValueError("`convolved` must be a boolean.")

    @property
    def low_loss(self):
        return self._low_loss

    @low_loss.setter
    def low_loss(self, value):
        if value is not None:
            if (
                value.axes_manager.navigation_shape
                != self.signal.axes_manager.navigation_shape
            ):
                raise ValueError(
                    "The signal does not have the same navigation dimension "
                    "as the signal it will be convolved with."
                )
            if not value.axes_manager.signal_axes[0].is_uniform:
                raise ValueError(
                    "Convolution is not supported with non-uniform signal axes."
                )
            self._low_loss = value
            self.set_convolution_axis()
            self.convolved = True
        else:
            self._low_loss = value
            self.convolution_axis = None
            self.convolved = False

    # Extend the list methods to call the _touch when the model is modified

    def set_convolution_axis(self):
        """
        Creates an axis to use to generate the data of the model in the precise
        scale to obtain the correct axis and origin after convolution.
        """
        ll_axis = self.low_loss.axes_manager.signal_axes[0]
        dimension = self.axis.size + ll_axis.size - 1
        step = self.axis.scale
        knot_position = ll_axis.size - ll_axis.value2index(0) - 1
        self.convolution_axis = generate_uniform_axis(
            self.axis.offset, step, dimension, knot_position
        )

    def append(self, component):
        """Append component to EELS model.

        Parameters
        ----------
        component
            HyperSpy component1D object.

        Raises
        ------
        NotImplementedError
            If the signal axis is a non-uniform axis.
        """
        super().append(component)
        if isinstance(component, EELSCLEdge):
            # Test that signal axis is uniform
            if not self.axes_manager[-1].is_uniform:
                raise NotImplementedError(
                    "This operation is not yet implemented "
                    "for non-uniform energy axes"
                )
            tem = self.signal.metadata.Acquisition_instrument.TEM
            component.set_microscope_parameters(
                E0=tem.beam_energy,
                alpha=tem.convergence_angle,
                beta=tem.Detector.EELS.collection_angle,
                energy_scale=self.axis.scale,
            )
            component.energy_scale = self.axis.scale
            component._set_fine_structure_coeff()
        self._classify_components()

    append.__doc__ = Model1D.append.__doc__

    def remove(self, component):
        super().remove(component)
        self._classify_components()

    remove.__doc__ = Model1D.remove.__doc__

    def _classify_components(self):
        """Classify components between background and ionization edge
        components.

        This method should be called every time that components are added and
        removed. An ionization edge becomes background when its onset falls to
        the left of the first non-masked energy channel. The ionization edges
        are stored in a list in the `edges` attribute. They are sorted by
        increasing `onset_energy`. The background components are stored in
        `_background_components`.

        """
        self.edges = []
        self._background_components = []
        for component in self:
            if isinstance(component, EELSCLEdge):
                if (
                    component.onset_energy.value
                    < self.axis.axis[self._channel_switches][0]
                ):
                    component.isbackground = True
                if component.isbackground is not True:
                    self.edges.append(component)
                else:
                    component.fine_structure_active = False
                    component.fine_structure_coeff.free = False
            elif isinstance(component, PowerLaw) or component.isbackground is True:
                self._background_components.append(component)

        if self.edges:
            self.edges.sort(key=EELSCLEdge._onset_energy)
            self.resolve_fine_structure()
        if len(self._background_components) > 1:
            self._backgroundtype = "mix"
        elif len(self._background_components) == 1:
            self._backgroundtype = self._background_components[0].__repr__()
            bg = self._background_components[0]
            if isinstance(bg, PowerLaw) and self.edges and not bg.A.map["is_set"].any():
                self.two_area_background_estimation()

    @property
    def _active_edges(self):
        return [edge for edge in self.edges if edge.active]

    @property
    def _active_background_components(self):
        return [bc for bc in self._background_components if bc.active]

    def _add_edges_from_subshells_names(self, e_shells=None):
        """Create the Edge instances and configure them appropriately

        Parameters
        ----------
        e_shells : list of strings
        """
        if self.signal._are_microscope_parameters_missing():
            raise ValueError(
                "The required microscope parameters are not defined in "
                "the EELS spectrum signal metadata. Use "
                "``set_microscope_parameters`` to set them."
            )
        if e_shells is None:
            e_shells = list(self.signal.subshells)
        e_shells.sort()
        master_edge = EELSCLEdge(e_shells.pop(), self.GOS)
        # If self.GOS was None, the GOS is set by eels_cl_edge so
        # we reassing the value of self.GOS
        self.GOS = master_edge.GOS._name
        self.append(master_edge)
        element = master_edge.element
        while len(e_shells) > 0:
            next_element = e_shells[-1].split("_")[0]
            if next_element != element:
                # New master edge
                self._add_edges_from_subshells_names(e_shells=e_shells)
            elif self.GOS == "hydrogenic":
                # The hydrogenic GOS includes all the L subshells in one
                # so we get rid of the others
                e_shells.pop()
            else:
                # Add the other subshells of the same element
                # and couple their intensity and onset_energy to that of the
                # master edge
                edge = EELSCLEdge(e_shells.pop(), GOS=self.GOS)

                edge.intensity.twin = master_edge.intensity
                edge.onset_energy.twin = master_edge.onset_energy
                edge.onset_energy.twin_function_expr = "x + {}".format(
                    (edge.GOS.onset_energy - master_edge.GOS.onset_energy)
                )
                edge.free_onset_energy = False
                self.append(edge)

    def resolve_fine_structure(self, preedge_safe_window_width=2, i1=0):
        """Adjust the fine structure of all edges to avoid overlapping

        This function is called automatically every time the position of an edge
        changes

        Parameters
        ----------
        preedge_safe_window_width : float
            minimum distance between the fine structure of an ionization edge
            and that of the following one. Default 2 (eV).

        """

        if self._suspend_auto_fine_structure_width is True:
            return

        if not self._active_edges:
            return

        while (
            self._active_edges[i1].fine_structure_active is False
            and i1 < len(self._active_edges) - 1
        ):
            i1 += 1
        if i1 < len(self._active_edges) - 1:
            i2 = i1 + 1
            while (
                self._active_edges[i2].fine_structure_active is False
                and i2 < len(self._active_edges) - 1
            ):
                i2 += 1
            if self._active_edges[i2].fine_structure_active is True:
                distance_between_edges = (
                    self._active_edges[i2].onset_energy.value
                    - self._active_edges[i1].onset_energy.value
                )
                if (
                    self._active_edges[i1].fine_structure_width
                    > distance_between_edges - self._preedge_safe_window_width
                ):
                    min_d = self._min_distance_between_edges_for_fine_structure
                    if (
                        distance_between_edges - self._preedge_safe_window_width
                    ) <= min_d:
                        _logger.info(
                            (
                                "Automatically deactivating the fine structure "
                                "of edge number %d to avoid conflicts with edge "
                                "number %d"
                            )
                            % (i2 + 1, i1 + 1)
                        )
                        self._active_edges[i2].fine_structure_active = False
                        self._active_edges[i2].fine_structure_coeff.free = False
                        self.resolve_fine_structure(i1=i2)
                    else:
                        new_fine_structure_width = (
                            distance_between_edges - self._preedge_safe_window_width
                        )
                        _logger.info(
                            (
                                "Automatically changing the fine structure "
                                "width of edge %d from %s eV to %s eV to avoid "
                                "conflicts with edge number %d"
                            )
                            % (
                                i1 + 1,
                                self._active_edges[i1].fine_structure_width,
                                new_fine_structure_width,
                                i2 + 1,
                            )
                        )
                        self._active_edges[
                            i1
                        ].fine_structure_width = new_fine_structure_width
                        self.resolve_fine_structure(i1=i2)
                else:
                    self.resolve_fine_structure(i1=i2)
        else:
            return

    def fit(self, kind="std", **kwargs):
        """Fits the model to the experimental data.

        Read more in the :ref:`User Guide <model.fitting>`.

        Parameters
        ----------
        kind : {"std", "smart"}, default "std"
            If "std", performs standard fit. If "smart",
            performs a smart_fit - for more details see
            the :ref:`User Guide <eels.fitting>`.
        %s

        Returns
        -------
        None

        See Also
        --------
        * :py:meth:`~hyperspy.model.BaseModel.fit`
        * :py:meth:`~hyperspy.model.BaseModel.multifit`
        * :py:meth:`~hyperspy.model.EELSModel.smart_fit`

        """
        if kind not in ["smart", "std"]:
            raise ValueError(f"kind must be either 'std' or 'smart', not '{kind}'")
        elif kind == "smart":
            return self.smart_fit(**kwargs)
        elif kind == "std":
            return Model1D.fit(self, **kwargs)

    fit.__doc__ %= FIT_PARAMETERS_ARG

    def smart_fit(self, start_energy=None, **kwargs):
        """Fits EELS edges in a cascade style.

        The fitting procedure acts in iterative manner along
        the energy-loss-axis. First it fits only the background
        up to the first edge. It continues by deactivating all
        edges except the first one, then performs the fit. Then
        it only activates the first two, fits, and repeats
        this until all edges are fitted simultaneously.

        Other, non-EELSCLEdge components, are never deactivated,
        and fitted on every iteration.

        Parameters
        ----------
        start_energy : {float, None}
            If float, limit the range of energies from the left to the
            given value.
        %s

        See Also
        --------
        * :py:meth:`~hyperspy.model.BaseModel.fit`
        * :py:meth:`~hyperspy.model.BaseModel.multifit`
        * :py:meth:`~hyperspy.model.EELSModel.fit`

        """
        cm = self.suspend_update if self._plot_active else dummy_context_manager
        with cm(update_on_resume=True):
            # Fit background
            self.fit_background(start_energy, **kwargs)

            # Fit the edges
            for i in range(0, len(self._active_edges)):
                self._fit_edge(i, start_energy, **kwargs)

    smart_fit.__doc__ %= FIT_PARAMETERS_ARG

    def _get_first_ionization_edge_energy(self, start_energy=None):
        """Calculate the first ionization edge energy.

        Returns
        -------
        iee : float or None
            The first ionization edge energy or None if no edge is defined in
            the model.

        """
        if not self._active_edges:
            return None
        start_energy = self._get_start_energy(start_energy)
        iee_list = [
            edge.onset_energy.value
            for edge in self._active_edges
            if edge.onset_energy.value > start_energy
        ]
        iee = min(iee_list) if iee_list else None
        return iee

    def _get_start_energy(self, start_energy=None):
        E0 = self.axis.axis[self._channel_switches][0]
        if not start_energy or start_energy < E0:
            start_energy = E0
        return start_energy

    def fit_background(self, start_energy=None, only_current=True, **kwargs):
        """Fit the background to the first active ionization edge
        in the energy range.

        Parameters
        ----------
        start_energy : {float, None}, optional
            If float, limit the range of energies from the left to the
            given value. Default None.
        only_current : bool, optional
            If True, only fit the background at the current coordinates.
            Default True.
        **kwargs : extra key word arguments
            All extra key word arguments are passed to fit or
            multifit.

        """

        # If there is no active background component do nothing
        if not self._active_background_components:
            return
        iee = self._get_first_ionization_edge_energy(start_energy=start_energy)
        if iee is not None:
            to_disable = [
                edge for edge in self._active_edges if edge.onset_energy.value >= iee
            ]
            E2 = iee - self._preedge_safe_window_width
            self.disable_edges(to_disable)
        else:
            E2 = None
        self.set_signal_range(start_energy, E2)
        if only_current:
            self.fit(**kwargs)
        else:
            self.multifit(**kwargs)
        self._channel_switches = copy.copy(self._backup_channel_switches)
        if iee is not None:
            self.enable_edges(to_disable)

    def two_area_background_estimation(self, E1=None, E2=None, powerlaw=None):
        """Estimates the parameters of a power law background with the two
        area method.

        Parameters
        ----------
        E1 : float
        E2 : float
        powerlaw : PowerLaw component or None
            If None, it will try to guess the right component from the
            background components of the model

        """
        if powerlaw is None:
            for component in self._active_background_components:
                if isinstance(component, components1d.PowerLaw):
                    if powerlaw is None:
                        powerlaw = component
                    else:
                        _logger.warning(
                            "There are more than two power law "
                            "background components defined in this model, "
                            "please use the powerlaw keyword to specify one"
                            " of them"
                        )
                        return
                else:  # No power law component
                    return

        ea = self.axis.axis[self._channel_switches]
        E1 = self._get_start_energy(E1)
        if E2 is None:
            E2 = self._get_first_ionization_edge_energy(start_energy=E1)
            if E2 is None:
                E2 = ea[-1]
            else:
                E2 = E2 - self._preedge_safe_window_width

        if not powerlaw.estimate_parameters(self.signal, E1, E2, only_current=False):
            _logger.warning(
                "The power law background parameters could not "
                "be estimated.\n"
                "Try choosing a different energy range for the estimation"
            )
            return

    def _fit_edge(self, edgenumber, start_energy=None, **kwargs):
        backup_channel_switches = self._channel_switches.copy()
        ea = self.axis.axis[self._channel_switches]
        if start_energy is None:
            start_energy = ea[0]
        # Declare variables
        active_edges = self._active_edges
        edge = active_edges[edgenumber]
        if (
            edge.intensity.twin is not None
            or edge.active is False
            or edge.onset_energy.value < start_energy
            or edge.onset_energy.value > ea[-1]
        ):
            return 1
        # Fitting edge 'edge.name'
        last_index = len(self._active_edges) - 1  # Last edge index
        i = 1
        twins = []
        # find twins
        while edgenumber + i <= last_index and (
            active_edges[edgenumber + i].intensity.twin is not None
            or active_edges[edgenumber + i].active is False
        ):
            if active_edges[edgenumber + i].intensity.twin is not None:
                twins.append(self._active_edges[edgenumber + i])
            i += 1
        if (edgenumber + i) > last_index:
            nextedgeenergy = ea[-1]
        else:
            nextedgeenergy = (
                active_edges[edgenumber + i].onset_energy.value
                - self._preedge_safe_window_width
            )

        # Backup the fsstate
        to_activate_fs = []
        for edge_ in [
            edge,
        ] + twins:
            if (
                edge_.fine_structure_active is True
                and edge_.fine_structure_coeff.free is True
                or edge_.fine_structure_components
            ):
                to_activate_fs.append(edge_)
        self.disable_fine_structure(to_activate_fs)

        # Smart Fitting

        # Without fine structure to determine onset_energy
        edges_to_activate = []
        for edge_ in self._active_edges[edgenumber + 1 :]:
            if edge_.active is True and edge_.onset_energy.value >= nextedgeenergy:
                edge_.active = False
                edges_to_activate.append(edge_)

        self.set_signal_range(start_energy, nextedgeenergy)
        if edge.free_onset_energy is True:
            edge.onset_energy.free = True
            self.fit(**kwargs)
            edge.onset_energy.free = False
            _logger.info("onset_energy = %s", edge.onset_energy.value)
            self._classify_components()
        elif edge.intensity.free is True:
            self.enable_fine_structure(to_activate_fs)
            self.remove_fine_structure_data(to_activate_fs)
            self.disable_fine_structure(to_activate_fs)
            self.fit(**kwargs)

        if len(to_activate_fs) > 0:
            self.set_signal_range(start_energy, nextedgeenergy)
            self.enable_fine_structure(to_activate_fs)
            self.fit(**kwargs)

        self.enable_edges(edges_to_activate)
        # Recover the _channel_switches. Remove it or make it smarter.
        self._channel_switches = backup_channel_switches

    def quantify(self):
        """Prints the value of the intensity of all the independent
        active EELS core loss edges defined in the model

        """
        elements = {}
        for edge in self._active_edges:
            if edge.active and edge.intensity.twin is None:
                element = edge.element
                subshell = edge.subshell
                if element not in elements:
                    elements[element] = {}
                elements[element][subshell] = edge.intensity.value
        print()
        print("Absolute quantification:")
        print("Elem.\tIntensity")
        for element in elements:
            if len(elements[element]) == 1:
                for subshell in elements[element]:
                    print("%s\t%f" % (element, elements[element][subshell]))
            else:
                for subshell in elements[element]:
                    print(
                        "%s_%s\t%f" % (element, subshell, elements[element][subshell])
                    )

    def remove_fine_structure_data(self, edges_list=None):
        """Remove the fine structure data from the fitting routine as
        defined in the fine_structure_width parameter of the
        component.EELSCLEdge

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False and edge.fine_structure_active is True:
                start = edge.onset_energy.value
                stop = start + edge.fine_structure_width
                self.remove_signal_range(start, stop)

    def enable_edges(self, edges_list=None):
        """Enable the edges listed in edges_list. If edges_list is
        None (default) all the edges with onset in the spectrum energy
        region will be enabled.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """

        if edges_list is None:
            edges_list = self.edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.active = True
        self.resolve_fine_structure()

    def disable_edges(self, edges_list=None):
        """Disable the edges listed in edges_list. If edges_list is None (default)
        all the edges with onset in the spectrum energy region will be
        disabled.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.active = False
        self.resolve_fine_structure()

    def enable_background(self):
        """Enable the background components."""
        for component in self._background_components:
            component.active = True

    def disable_background(self):
        """Disable the background components."""
        for component in self._active_background_components:
            component.active = False

    def enable_fine_structure(self, edges_list=None):
        """Enable the fine structure of the edges listed in edges_list.

        If edges_list is None (default) the fine structure of all the edges
        with onset in the spectrum energy region will be enabled.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.fine_structure_active = True
        self.resolve_fine_structure()

    def disable_fine_structure(self, edges_list=None):
        """Disable the fine structure of the edges listed in edges_list.
        If edges_list is None (default) the fine structure of all the edges
        with onset in the spectrum energy region will be disabled.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.fine_structure_active = False
        self.resolve_fine_structure()

    def set_all_edges_intensities_positive(self):
        """
        Set all edges intensities positive by setting ``ext_force_positive``
        and ``ext_bounded`` to ``True``.

        Returns
        -------
        None.

        """
        for edge in self._active_edges:
            edge.intensity.ext_force_positive = True
            edge.intensity.ext_bounded = True

    def unset_all_edges_intensities_positive(self):
        """
        Unset all edges intensities positive by setting ``ext_force_positive``
        and ``ext_bounded`` to ``False``.

        Returns
        -------
        None.

        """
        for edge in self._active_edges:
            edge.intensity.ext_force_positive = False
            edge.intensity.ext_bounded = False

    def enable_free_onset_energy(self, edges_list=None):
        """Enable the automatic freeing of the onset_energy parameter during a
        smart fit for the edges listed in edges_list.
        If edges_list is None (default) the onset_energy of all the edges
        with onset in the spectrum energy region will be freed.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.free_onset_energy = True

    def disable_free_onset_energy(self, edges_list=None):
        """Disable the automatic freeing of the onset_energy parameter during a
        smart fit for the edges listed in edges_list.
        If edges_list is None (default) the onset_energy of all the edges
        with onset in the spectrum energy region will not be freed.
        Note that if their attribute edge.onset_energy.free is True, the
        parameter will be free during the smart fit.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """

        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.free_onset_energy = True

    def fix_edges(self, edges_list=None):
        """Fixes all the parameters of the edges given in edges_list.
        If edges_list is None (default) all the edges will be fixed.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.intensity.free = False
                edge.onset_energy.free = False
                edge.fine_structure_coeff.free = False

    def free_edges(self, edges_list=None):
        """Frees all the parameters of the edges given in edges_list.
        If edges_list is None (default) all the edges will be freed.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """

        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.set_parameters_free(
                    parameter_name_list=[
                        "intensity",
                        "onset_energy",
                        "fine_structure_coeff",
                    ]
                )

    def fix_fine_structure(self, edges_list=None):
        """Fixes the fine structure of the edges given in edges_list.
        If edges_list is None (default) all the edges will be fixed.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.fix_fine_structure()

    def free_fine_structure(self, edges_list=None):
        """Frees the fine structure of the edges given in edges_list.
        If edges_list is None (default) all the edges will be freed.

        Parameters
        ----------
        edges_list : None or list of EELSCLEdge or list of edge names
            If None, the operation is performed on all the edges in the model.
            Otherwise, it will be performed only on the listed components.

        See Also
        --------
        enable_edges, disable_edges, enable_background,
        disable_background, enable_fine_structure,
        disable_fine_structure, set_all_edges_intensities_positive,
        unset_all_edges_intensities_positive, enable_free_onset_energy,
        disable_free_onset_energy, fix_edges, free_edges, fix_fine_structure,
        free_fine_structure

        """
        if edges_list is None:
            edges_list = self._active_edges
        else:
            edges_list = [self._get_component(x) for x in edges_list]
        for edge in edges_list:
            if edge.isbackground is False:
                edge.free_fine_structure()

    def suspend_auto_fine_structure_width(self):
        """Disable the automatic adjustment of the core-loss edges fine
        structure width.

        See Also
        --------
        resume_auto_fine_structure_width

        """
        if self._suspend_auto_fine_structure_width is False:
            self._suspend_auto_fine_structure_width = True
        else:
            warnings.warn("Already suspended, does nothing.")

    def resume_auto_fine_structure_width(self, update=True):
        """Enable the automatic adjustment of the core-loss edges fine
        structure width.

        Parameters
        ----------
        update : bool, optional
            If True, also execute the automatic adjustment (default).

        See Also
        --------
        suspend_auto_fine_structure_width

        """
        if self._suspend_auto_fine_structure_width is True:
            self._suspend_auto_fine_structure_width = False
            if update is True:
                self.resolve_fine_structure()
        else:
            warnings.warn("Not suspended, nothing to resume.")



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np

import hyperspy.api as hs


class VolumePlasmonDrude(hs.model.components1D.Expression):
    r"""
    Drude volume plasmon energy loss function component, the energy loss
    function is defined as:

    .. math::

       f(E) = I_0 \frac{E(\Delta E_p)E_p^2}{(E^2-E_p^2)^2+(E\Delta E_p)^2}

    ================== ===============
    Variable            Parameter
    ================== ===============
    :math:`I_0`         intensity
    :math:`E_p`         plasmon_energy
    :math:`\Delta E_p`  fwhm
    ================== ===============

    Parameters
    ----------
    intensity : float
    plasmon_energy : float
    fwhm : float
    **kwargs
        Extra keyword arguments are passed to the
        :py:class:`hyperspy._components.expression.Expression` component.

    Notes
    -----
    Refer to Egerton, R. F., Electron Energy-Loss Spectroscopy in the
    Electron Microscope, 2nd edition, Plenum Press 1996, pp. 154-158
    for details, including original equations.
    """

    def __init__(
        self,
        intensity=1.0,
        plasmon_energy=15.0,
        fwhm=1.5,
        module="numexpr",
        compute_gradients=False,
        **kwargs,
    ):
        super().__init__(
            expression="where(x > 0, intensity * (pe2 * x * fwhm) \
                        / ((x ** 2 - pe2) ** 2 + (x * fwhm) ** 2), 0); \
                        pe2 = plasmon_energy ** 2",
            name="VolumePlasmonDrude",
            intensity=intensity,
            plasmon_energy=plasmon_energy,
            fwhm=fwhm,
            position="plasmon_energy",
            module=module,
            autodoc=False,
            compute_gradients=compute_gradients,
            linear_parameter_list=["intensity"],
            check_parameter_linearity=False,
            **kwargs,
        )

    # Partial derivative with respect to the plasmon energy E_p
    def grad_plasmon_energy(self, x):
        plasmon_energy = self.plasmon_energy.value
        fwhm = self.fwhm.value
        intensity = self.intensity.value

        return np.where(
            x > 0,
            2
            * x
            * fwhm
            * plasmon_energy
            * intensity
            * (
                (x**4 + (x * fwhm) ** 2 - plasmon_energy**4)
                / (
                    x**4
                    + x**2 * (fwhm**2 - 2 * plasmon_energy**2)
                    + plasmon_energy**4
                )
                ** 2
            ),
            0,
        )

    # Partial derivative with respect to the plasmon linewidth delta_E_p
    def grad_fwhm(self, x):
        plasmon_energy = self.plasmon_energy.value
        fwhm = self.fwhm.value
        intensity = self.intensity.value

        return np.where(
            x > 0,
            x
            * plasmon_energy
            * intensity
            * (
                (
                    x**4
                    - x**2 * (2 * plasmon_energy**2 + fwhm**2)
                    + plasmon_energy**4
                )
                / (
                    x**4
                    + x**2 * (fwhm**2 - 2 * plasmon_energy**2)
                    + plasmon_energy**4
                )
                ** 2
            ),
            0,
        )

    def grad_intensity(self, x):
        return self.function(x) / self.intensity.value



"""
Simple arctan fit
=================

Fit an arctan function.

"""

import numpy as np
import hyperspy.api as hs

# Generate the data and make the spectrum
data = np.arctan(np.arange(-500, 500))
s = hs.signals.Signal1D(data)
s.axes_manager[0].offset = -500
s.axes_manager[0].units = ""
s.axes_manager[0].name = "x"
s.metadata.General.title = "Simple arctan fit"
s.set_signal_origin("simulation")

s.add_gaussian_noise(0.1)

# Make the arctan component for use in the model
arctan_component = hs.model.components1D.Arctan()

# Create the model and add the arctan component
m = s.create_model()
m.append(arctan_component)

# Fit the arctan component to the spectrum
m.fit()

# Print the result of the fit
m.print_current_values()

# Plot the spectrum and the model fitting
m.plot()



"""
EELS curve fitting
==================

Performs curve fitting to an EELS spectrum, plots the result and saves it as
png file.

"""

import hyperspy.api as hs

s = hs.load("coreloss_spectrum.msa", signal_type="EELS")
ll = hs.load("lowloss_spectrum.msa", signal_type="EELS")

s.add_elements(("Mn", "O"))
s.set_microscope_parameters(
    beam_energy=300, convergence_angle=24.6, collection_angle=13.6
)

m = s.create_model(low_loss=ll)
m.enable_fine_structure()
m.multifit(kind="smart")
m.plot()



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np

from hyperspy.docstrings.parameters import FUNCTION_ND_DOCSTRING
from hyperspy._components.expression import Expression


class DoublePowerLaw(Expression):
    r"""Double power law component for EELS spectra.

    .. math::

        f(x) = A \cdot [s_r \cdot (x - x_0 - x_s)^{-r} + (x - x_0)^{-r}]

    ============= =============
     Variable      Parameter
    ============= =============
     :math:`A`     A
     :math:`r`     r
     :math:`x_0`   origin
     :math:`x_s`   shift
     :math:`s_r`   ratio
    ============= =============

    Parameters
    ----------
    A : float
        Height parameter.
    r : float
        Power law coefficient.
    origin : float
        Location parameter.
    shift : float
        Offset of second power law.
    ratio : float
        Height ratio of the two power law components.
    **kwargs
        Extra keyword arguments are passed to the
        :py:class:`~._components.expression.Expression` component.

    Attributes
    ----------
    left_cutoff : float
        For x <= left_cutoff, the function returns 0. Default value is 0.0.
    """

    def __init__(
        self,
        A=1e-5,
        r=3.0,
        origin=0.0,
        shift=20.0,
        ratio=1.0,
        left_cutoff=0.0,
        module="numexpr",
        compute_gradients=False,
        **kwargs,
    ):
        super().__init__(
            expression="where(x > left_cutoff, \
                        A * (ratio * (x - origin - shift) ** -r \
                        + (x - origin) ** -r), 0)",
            name="DoublePowerLaw",
            A=A,
            r=r,
            origin=origin,
            shift=shift,
            ratio=ratio,
            left_cutoff=left_cutoff,
            position="origin",
            autodoc=False,
            module=module,
            compute_gradients=compute_gradients,
            linear_parameter_list=["A"],
            check_parameter_linearity=False,
            **kwargs,
        )

        # Boundaries
        self.A.bmin = 0.0
        self.A.bmax = None
        self.r.bmin = 1.0
        self.r.bmax = 5.0

        self.isbackground = True
        self.convolved = True

    def function_nd(self, axis):
        """%s"""
        return super().function_nd(axis)

    function_nd.__doc__ %= FUNCTION_ND_DOCSTRING

    # Define gradients
    def grad_A(self, x):
        return self.function(x) / self.A.value

    def grad_r(self, x):
        return np.where(
            x > self.left_cutoff.value,
            -self.A.value
            * self.ratio.value
            * (x - self.origin.value - self.shift.value) ** (-self.r.value)
            * np.log(x - self.origin.value - self.shift.value)
            - self.A.value
            * (x - self.origin.value) ** (-self.r.value)
            * np.log(x - self.origin.value),
            0,
        )

    def grad_origin(self, x):
        return np.where(
            x > self.left_cutoff.value,
            self.A.value
            * self.r.value
            * self.ratio.value
            * (x - self.origin.value - self.shift.value) ** (-self.r.value - 1)
            + self.A.value
            * self.r.value
            * (x - self.origin.value) ** (-self.r.value - 1),
            0,
        )

    def grad_shift(self, x):
        return np.where(
            x > self.left_cutoff.value,
            self.A.value
            * self.r.value
            * self.ratio.value
            * (x - self.origin.value - self.shift.value) ** (-self.r.value - 1),
            0,
        )

    def grad_ratio(self, x):
        return np.where(
            x > self.left_cutoff.value,
            self.A.value
            * (x - self.origin.value - self.shift.value) ** (-self.r.value),
            0,
        )



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The HyperSpy developers
#
# This file is part of HyperSpy.
#
# HyperSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# HyperSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with HyperSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import math

from hyperspy.component import Component, _get_scaling_factor
from hyperspy._components.gaussian import _estimate_gaussian_parameters


sqrt2pi = math.sqrt(2 * math.pi)
sigma2fwhm = 2 * math.sqrt(2 * math.log(2))


def voigt(x, FWHM=1, gamma=1, center=0, scale=1):
    r"""Voigt lineshape.

    The voigt peak is the convolution of a Lorentz peak with a Gaussian peak:

    .. math::
        f(x) = G(x) \cdot L(x)

    where :math:`G(x)` is the Gaussian function and :math:`L(x)` is the
    Lorentzian function. In this case using an approximate formula by David
    (see Notes). This approximation improves on the pseudo-Voigt function
    (linear combination instead of convolution of the distributions) and is,
    to a very good approximation, equivalent to a Voigt function:

    .. math::
        z(x) &= \frac{x + i \gamma}{\sqrt{2} \sigma} \\
        w(z) &= \frac{e^{-z^2} \text{erfc}(-i z)}{\sqrt{2 \pi} \sigma} \\
        f(x) &= A \cdot \Re\left\{ w \left[ z(x - x_0) \right] \right\}


    ============== =============
    Variable        Parameter
    ============== =============
    :math:`x_0`     center
    :math:`A`       scale
    :math:`\gamma`  gamma
    :math:`\sigma`  sigma
    ============== =============


    Parameters
    ----------
    gamma : real
       The half-width half-maximum of the Lorentzian.
    FWHM : real
       The FWHM = :math:`2 \sigma \sqrt{(2 \log(2))}` of the Gaussian.
    center : real
       Location of the center of the peak.
    scale : real
       Value at the highest point of the peak.

    Notes
    -----
    Ref: W.I.F. David, J. Appl. Cryst. (1986). 19, 63-64
    doi:10.1107/S0021889886089999
    """
    # wofz function = w(z) = Fad[d][e][y]eva function = exp(-z**2)erfc(-iz)
    from scipy.special import wofz

    sigma = FWHM / 2.3548200450309493
    z = (np.asarray(x) - center + 1j * gamma) / (sigma * math.sqrt(2))
    V = wofz(z) / (math.sqrt(2 * np.pi) * sigma)
    return scale * V.real


class PESVoigt(Component):
    r"""Voigt component for photoemission spectroscopy data analysis.

    Voigt profile component with support for shirley background,
    non_isochromaticity, transmission_function corrections and spin orbit
    splitting specially suited for photoemission spectroscopy data analysis.

    .. math::
        f(x) = G(x) \cdot L(x)

    where :math:`G(x)` is the Gaussian function and :math:`L(x)` is the
    Lorentzian function. This component uses an approximate formula by David
    (see Notes).


    Parameters
    ----------

    area : Parameter
        Intensity below the peak.
    centre: Parameter
        Location of the maximum of the peak.
    FWHM : Parameter
        FWHM = :math:`2 \sigma \sqrt{(2 \log(2))}` of the Gaussian distribution.
    gamma : Parameter
        :math:`\gamma` of the Lorentzian distribution.
    resolution : Parameter
    shirley_background : Parameter
    non_isochromaticity : Parameter
    transmission_function : Parameter
    spin_orbit_splitting : Bool
    spin_orbit_branching_ratio : float
    spin_orbit_splitting_energy : float

    Notes
    -----
    Uses an approximate formula according to
    W.I.F. David, J. Appl. Cryst. (1986). 19, 63-64.
    doi:10.1107/S0021889886089999
    """

    def __init__(self):
        Component.__init__(
            self,
            (
                "area",
                "centre",
                "FWHM",
                "gamma",
                "resolution",
                "shirley_background",
                "non_isochromaticity",
                "transmission_function",
            ),
        )
        self._position = self.centre
        self.FWHM.value = 1
        self.gamma.value = 0
        self.area.value = 1
        self.resolution.value = 0
        self.resolution.free = False
        self.shirley_background.free = False
        self.non_isochromaticity.value = 0
        self.non_isochromaticity.free = False
        self.transmission_function.value = 1
        self.transmission_function.free = False
        # Options
        self.shirley_background.active = False
        self.spin_orbit_splitting = False
        self.spin_orbit_branching_ratio = 0.5
        self.spin_orbit_splitting_energy = 0.61
        self.isbackground = False
        self.convolved = True

    def function(self, x):
        area = self.area.value * self.transmission_function.value
        centre = self.centre.value
        ab = self.non_isochromaticity.value
        if self.resolution.value == 0:
            FWHM = self.FWHM.value
        else:
            FWHM = math.sqrt(self.FWHM.value**2 + self.resolution.value**2)
        gamma = self.gamma.value
        k = self.shirley_background.value
        f = voigt(x, FWHM=FWHM, gamma=gamma, center=centre - ab, scale=area)
        if self.spin_orbit_splitting:
            ratio = self.spin_orbit_branching_ratio
            shift = self.spin_orbit_splitting_energy
            f2 = voigt(
                x,
                FWHM=FWHM,
                gamma=gamma,
                center=centre - ab - shift,
                scale=area * ratio,
            )
            f += f2
        if self.shirley_background.active:
            cf = np.cumsum(f)
            cf = cf[-1] - cf
            self.cf = cf
            return cf * k + f
        else:
            return f

    def estimate_parameters(self, signal, E1, E2, only_current=False):
        """Estimate the Voigt function by calculating the momenta of the
        Gaussian.

        Parameters
        ----------
        signal : Signal1D instance
        x1 : float
            Defines the left limit of the spectral range to use for the
            estimation.
        x2 : float
            Defines the right limit of the spectral range to use for the
            estimation.

        only_current : bool
            If False estimates the parameters for the full dataset.

        Returns
        -------
         : bool
            Exit status required for the
            :py:meth:`~._signals.signal1d.Signal1D.remove_background` function.

        Notes
        -----
        Adapted from https://scipy-cookbook.readthedocs.io/items/FittingData.html

        Examples
        --------

        >>> g = hs.model.components1D.PESVoigt()
        >>> x = np.arange(-10, 10, 0.01)
        >>> data = np.zeros((32, 32, 2000))
        >>> data[:] = g.function(x).reshape((1, 1, 2000))
        >>> s = hs.signals.Signal1D(data)
        >>> s.axes_manager[-1].offset = -10
        >>> s.axes_manager[-1].scale = 0.01
        >>> g.estimate_parameters(s, -10, 10, False)

        """
        super()._estimate_parameters(signal)
        axis = signal.axes_manager.signal_axes[0]
        centre, height, sigma = _estimate_gaussian_parameters(
            signal, E1, E2, only_current
        )
        scaling_factor = _get_scaling_factor(signal, axis, centre)

        if only_current is True:
            self.centre.value = centre
            self.FWHM.value = sigma * sigma2fwhm
            self.area.value = height * sigma * sqrt2pi
            if axis.is_binned:
                self.area.value /= scaling_factor
            return True
        else:
            if self.area.map is None:
                self._create_arrays()
            self.area.map["values"][:] = height * sigma * sqrt2pi
            if axis.is_binned:
                self.area.map["values"][:] /= scaling_factor
            self.area.map["is_set"][:] = True
            self.FWHM.map["values"][:] = sigma * sigma2fwhm
            self.FWHM.map["is_set"][:] = True
            self.centre.map["values"][:] = centre
            self.centre.map["is_set"][:] = True
            self.fetch_stored_values()
            return True



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

import numpy as np
import pytest

import exspy as ex
from exspy.misc.elements import elements_db
import hyperspy.api as hs


class TestWeightToFromAtomic:
    def setup_method(self, method):
        # TiO2
        self.elements = ("Ti", "O")
        natoms = (1, 2)
        self.at = [100 * nat / float(sum(natoms)) for nat in natoms]
        atomic_weight = np.array(
            [
                elements_db[element].General_properties.atomic_weight
                for element in self.elements
            ]
        )
        mol_weight = atomic_weight * natoms
        self.wt = [100 * w / mol_weight.sum() for w in mol_weight]

    def test_weight_to_atomic(self):
        cwt = ex.material.weight_to_atomic(self.wt, self.elements)
        np.testing.assert_allclose(cwt[0], self.at[0])
        np.testing.assert_allclose(cwt[1], self.at[1])

    def test_atomic_to_weight(self):
        cat = ex.material.atomic_to_weight(self.at, self.elements)
        np.testing.assert_allclose(cat[0], self.wt[0])
        np.testing.assert_allclose(cat[1], self.wt[1])

    def test_multi_dim(self):
        elements = ("Cu", "Sn")
        wt = np.array([[[88] * 2] * 3, [[12] * 2] * 3])
        at = ex.material.weight_to_atomic(wt, elements)
        np.testing.assert_allclose(
            at[:, 0, 0], np.array([93.196986, 6.803013]), atol=1e-3
        )
        wt2 = ex.material.atomic_to_weight(at, elements)
        np.testing.assert_allclose(wt, wt2)


def test_density_of_mixture():
    # Bronze
    elements = ("Cu", "Sn")
    wt = (88.0, 12.0)
    densities = np.array(
        [elements_db[element].Physical_properties.density_gcm3 for element in elements]
    )

    volumes = wt * densities
    density = volumes.sum() / 100.0
    np.testing.assert_allclose(
        density, ex.material.density_of_mixture(wt, elements, mean="weighted")
    )

    volumes = wt / densities
    density = 100.0 / volumes.sum()
    np.testing.assert_allclose(density, ex.material.density_of_mixture(wt, elements))

    wt = np.array([[[88] * 2] * 3, [[12] * 2] * 3])
    np.testing.assert_allclose(
        density, ex.material.density_of_mixture(wt, elements)[0, 0]
    )

    # Testing whether the correct exception is raised upon unknown density
    elements = ("Cu", "Sn", "At")
    wt = (87.0, 12.0, 1.0)
    with pytest.raises(ValueError):
        ex.material.density_of_mixture(wt, elements)


def test_mac():
    np.testing.assert_allclose(
        ex.material.mass_absorption_coefficient("Al", 3.5), 506.0153356472
    )
    np.testing.assert_allclose(
        ex.material.mass_absorption_coefficient("Ta", [1, 3.2, 2.3]),
        [3343.7083701143229, 1540.0819991890, 3011.264941118],
    )
    np.testing.assert_allclose(
        ex.material.mass_absorption_coefficient("Zn", "Zn_La"), 1413.291119134
    )
    np.testing.assert_allclose(
        ex.material.mass_absorption_coefficient("Zn", ["Cu_La", "Nb_La"]),
        [1704.7912903000029, 1881.2081950943339],
    )


def test_mixture_mac():
    np.testing.assert_allclose(
        ex.material.mass_absorption_mixture([50, 50], ["Al", "Zn"], "Al_Ka"),
        2587.4161643905127,
    )
    elements = ("Cu", "Sn")
    lines = [0.5, "Al_Ka"]
    wt = np.array([[[88.0] * 2] * 3, [[12.0] * 2] * 3])
    np.testing.assert_array_almost_equal(
        ex.material.mass_absorption_mixture(wt, elements, lines)[:, 0, 0],
        np.array([8003.05391481, 4213.4235561]),
    )
    wt = hs.signals.Signal2D(wt).split()
    mac = ex.material.mass_absorption_mixture(wt, elements, lines)
    np.testing.assert_array_almost_equal(mac[0].data[0, 0], 8003.053914)



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.


import numpy as np
from hyperspy.component import Component
from hyperspy._components.gaussian import Gaussian


class Vignetting(Component):

    """
    Model the vignetting of the lens with a cos^4 law multiplied by lines on
    the edges
    """

    def __init__(self):
        Component.__init__(
            self,
            [
                "optical_center",
                "height",
                "period",
                "left_slope",
                "right_slope",
                "left",
                "right",
                "sigma",
            ],
        )
        self.left.value = np.nan
        self.right.value = np.nan
        self.side_vignetting = False
        self.fix_side_vignetting()
        self.gaussian = Gaussian()
        self.gaussian.centre.free, self.gaussian.A.free = False, False
        self.sigma.value = 1.0
        self.gaussian.A.value = 1.0
        self.period.value = 1.0
        self.extension_nch = 100
        self._position = self.optical_center

    def function(self, x):
        sigma = self.sigma.value
        x0 = self.optical_center.value
        A = self.height.value
        period = self.period.value
        la = self.left_slope.value
        ra = self.right_slope.value
        l = self.left.value
        r = self.right.value
        ex = self.extension_nch
        if self.side_vignetting is True:
            x = x.tolist()
            x = (
                list(range(-ex, 0))
                + x
                + list(range(int(x[-1]) + 1, int(x[-1]) + ex + 1))
            )
            x = np.array(x)
            v1 = A * np.cos((x - x0) / (2 * np.pi * period)) ** 4
            v2 = np.where(
                x < l, 1.0 - (l - x) * la, np.where(x < r, 1.0, 1.0 - (x - r) * ra)
            )
            self.gaussian.sigma.value = sigma
            self.gaussian.origin.value = (x[-1] + x[0]) / 2
            result = np.convolve(self.gaussian.function(x), v1 * v2, "same")
            return result[ex:-ex]
        else:
            return A * np.cos((x - x0) / (2 * np.pi * period)) ** 4

    def free_side_vignetting(self):
        self.left.free = True
        self.right.free = True
        self.left_slope.free = True
        self.right_slope.free = True
        self.sigma.free = True

    def fix_side_vignetting(self):
        self.left.free = False
        self.right.free = False
        self.left_slope.free = False
        self.right_slope.free = False
        self.sigma.free = False

    def free_cos_vignetting(self):
        self.optical_center.free = True
        self.period.free = True
        self.height.free = True

    def fix_cos_vignetting(self):
        self.optical_center.free = False
        self.period.free = False
        self.height.free = False



# -*- coding: utf-8 -*-
# Copyright 2007-2023 The exSpy developers
#
# This file is part of exSpy.
#
# exSpy is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# exSpy is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with exSpy. If not, see <https://www.gnu.org/licenses/#GPL>.

from pathlib import Path

import h5py
import pooch
import pytest

from exspy._defaults_parser import preferences
from exspy.misc.eels.gosh_gos import GoshGOS
from exspy.misc.eels.hartree_slater_gos import HartreeSlaterGOS
from exspy.misc.eels import HydrogenicGOS
from exspy.misc.elements import elements


GOSH10 = pooch.retrieve(
    url="doi:10.5281/zenodo.6599071/Segger_Guzzinati_Kohl_1.0.0.gos",
    known_hash="md5:d65d5c23142532fde0a80e160ab51574",
    progressbar=False,
)


@pytest.mark.skipif(
    not Path(preferences.EELS.eels_gos_files_path).exists(),
    reason="Hartree-Slater GOS not available",
)
def test_hartree_slater_gos():
    gos = HartreeSlaterGOS("Ti_L3")
    gos.read_elements()


def test_hydrogenic_gos_error_M_shells():
    with pytest.raises(ValueError):
        _ = HydrogenicGOS("Ti_M2")


def test_element_not_in_database():
    with pytest.raises(ValueError):
        _ = HydrogenicGOS("Lr_M2")


def test_subshell_not_in_database():
    with pytest.raises(ValueError):
        _ = HydrogenicGOS("Ti_L4")


def test_gosh_not_in_conventions():
    gos = GoshGOS("Ti_L2")
    gos.subshell = "L234"
    with pytest.raises(ValueError):
        gos.read_gos_data()


def test_gosh_not_in_file():
    # Use version 1.0 which doesn't have the Ac element
    with pytest.raises(ValueError):
        _ = GoshGOS("Ac_L3", gos_file_path=GOSH10)


def test_binding_energy_database():
    gos = GoshGOS("Ti_L3")
    gosh15 = h5py.File(gos.gos_file_path)
    for element in gosh15.keys():
        # These elements are not in the database
        if element not in ["Bk", "Cf", "Cm", "metadata"]:
            assert "Binding_energies" in elements[element]["Atomic_properties"].keys()
